==> fsh-0.1.tar.gz: fsh-0.1/fsh.py <==
#  fsh - fast remote execution
#  Copyright (C) 1999 by Per Cederqvist.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

import fshlib
import os
import select
import socket
import string
import sys

class remote:
    def __init__(self, server, method, command):
        self.s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.s.connect("/tmp/fshd-%d/%s.%s" % (os.getuid(), server, method))
        self.sock = self.s.fileno()
        self.send_queue = ""
        self.receive_queue = ""
        self.send("new %s\n" % fshlib.hollerith(command))
        
    def select_set(self):
        w = []
        if self.sock == -1:
            return [], [], []
        if self.send_queue != "":
            w.append(self.sock)
        return [self.sock], w, []

    def select_action(self, r, w, e):
        if self.sock in r:
            try:
		data = os.read(self.sock, 4096)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		else:
		    raise
	    if data == "":
                self.s.close()
                self.sock = -1
	    else:
                self.receive_queue = self.receive_queue + data

        if self.sock in w:
            try:
                sz = os.write(self.sock, self.send_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		elif eno == errno.EPIPE:
		    self.s.close()
                    self.sock = -1
		    return
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
            #os.write(debug, "sent this: %s.\n" % self.send_queue[:sz])
	    self.send_queue = self.send_queue[sz:]

    def send(self, data):
	#os.write(debug, "enqueueing this: %s.\n" % data)
        self.send_queue = self.send_queue + data
        if self.sock == -1:
            self.send_queue = ""

    def poll_response(self):
        [cmd, data], self.receive_queue = fshlib.parse_line(
            self.receive_queue, 0)

        return [cmd, data]

    def eof_seen(self):
        return self.receive_queue == "" and self.sock == -1

    def close(self):
        self.s.close()
        self.sock = -1

    def drained(self):
        return self.sock == -1

class local_side:
    def __init__(self):
        self.stdin_quota = fshlib.QUOTA
        self.stdout_quota = fshlib.QUOTA
        self.stderr_quota = fshlib.QUOTA
        self.stdin_counter = 0
        self.stdout_counter = 0
        self.stderr_counter = 0
        self.stdout_queue = ""
        self.stderr_queue = ""
        self.stdin_fd = 0
        self.stdout_fd = 1
        self.stderr_fd = 2
        self.epipe_seen = 0
	self.stdout_pending_close = 0
	self.stderr_pending_close = 0

    def select_set(self):
        r = []
        w = []
        if self.stdin_fd != -1 and self.stdin_counter < self.stdin_quota:
            r.append(self.stdin_fd)
        if self.stdout_fd != -1 and self.stdout_queue != "":
            w.append(self.stdout_fd)
        if self.stderr_fd != -1 and self.stderr_queue != "":
            w.append(self.stderr_fd)
        return r, w, []

    def select_action(self, r, w, e, srv):
        if self.stdin_fd in r:
            wanted = min(4096, self.stdin_quota - self.stdin_counter)
            assert(wanted > 0)
            try:
		data = os.read(self.stdin_fd, wanted)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		else:
		    raise
	    if data == "":
                os.close(self.stdin_fd)
                self.stdin_fd = -1
                srv.send("eof-stdin\n")
	    else:
                srv.send("stdin %s\n" % fshlib.hollerith(data))
                self.stdin_counter = self.stdin_counter + len(data)

        if self.stdout_fd in w:
	    try:
		sz = os.write(self.stdout_fd, self.stdout_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		elif eno == errno.EPIPE:
                    srv.send("eof-stdout\n")
                    os.close(self.stdout_fd)
                    self.stdout_fd = -1
		    #os.write(debug, "epipe seen I\n")
                    self.epipe_seen = 1
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
	    self.stdout_queue = self.stdout_queue[sz:]
            self.stdout_counter = self.stdout_counter + sz
	    #os.write(debug, "successful write of %d bytes.\n" % sz)
            if self.stdout_pending_close and self.stdout_queue == "":
		#os.write(debug, "closing stdout due to pend.\n")
                os.close(self.stdout_fd)
                self.stdout_fd = -1
	    #os.write(debug, "hmm.\n")
            if self.stdout_fd != -1 and \
               (self.stdout_quota - self.stdout_counter) < fshlib.QUOTA/2:

		#os.write(debug, "increasing quota.\n")
                self.stdout_quota = self.stdout_quota + fshlib.QUOTA
                srv.send("stdout-flow %s\n" % fshlib.hollerith(
                    self.stdout_quota))
		#os.write(debug, "increased quota.\n")

        if self.stderr_fd in w:
	    try:
		sz = os.write(self.stderr_fd, self.stderr_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		elif eno == errno.EPIPE:
                    srv.send("eof-stderr\n")
                    os.close(self.stderr_fd)
                    self.stderr_fd = -1
		    #os.write(debug, "epipe seen II\n")
                    self.epipe_seen = 1
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
	    self.stderr_queue = self.stderr_queue[sz:]
            self.stderr_counter = self.stderr_counter + sz
            if self.stderr_pending_close and self.stderr_queue == "":
                os.close(self.stderr_fd)
                self.stderr_fd = -1
            if self.stderr_fd != -1 and \
               (self.stderr_quota - self.stderr_counter) < fshlib.QUOTA/2:

                self.stderr_quota = self.stderr_quota + fshlib.QUOTA
                srv.send("stderr-flow %s\n" % fshlib.hollerith(
                    self.stderr_quota))

    def stdout_send(self, data):
        self.stdout_queue = self.stdout_queue + data

    def stderr_send(self, data):
        self.stderr_queue = self.stderr_queue + data

    def eof_stdin(self):
        if self.stdin_fd != -1:
            os.close(self.stdin_fd)
            self.stdin_fd = -1

    def eof_stdout(self):
        self.stdout_pending_close = 1
	if self.stdout_queue == ""and self.stdout_fd != -1:
	    os.close(self.stdout_fd)
	    self.stdout_fd = -1

    def eof_stderr(self):
	#os.write(debug, "eof_stderr: start\n")
        self.stderr_pending_close = 1
	#os.write(debug, "eof_stderr: I\n")
	if self.stderr_queue == "" and self.stderr_fd != -1:
	    #os.write(debug, "eof_stderr: II (%d, %d)\n" % (debug, self.stderr_fd))
	    os.close(self.stderr_fd)
	    #os.write(debug, "eof_stderr: III\n")
	    self.stderr_fd = -1
	    #os.write(debug, "eof_stderr: IV\n")
	#os.write(debug, "eof_stderr: V\n")

    def stdin_flow(self, data):
        self.stdin_quota = data

    def drained(self):
        return self.stdout_fd == -1 and self.stderr_fd == -1

def docmd(server, method, command):
    rem = remote(server, method, command)
    loc = local_side()
    exitval = None
    while 1:
	#os.write(debug, "at top of main loop\n")
        while 1:
            cmd, data = rem.poll_response()
            if cmd == None:
                break
	    #os.write(debug, "got cmd: %s (%s)\n" % (cmd, data))
            if cmd == "stdout":
                loc.stdout_send(data)
            elif cmd == "stderr":
                loc.stderr_send(data)
            elif cmd == "eof-stdin":
                loc.eof_stdin()
            elif cmd == "eof-stdout":
                loc.eof_stdout()
            elif cmd == "eof-stderr":
                loc.eof_stderr()
            elif cmd == "exit":
                exitval = string.atoi(data)
            elif cmd == "signal-exit":
                # What do we do here?
                exitval = string.atoi(data)
            elif cmd == "eos":
		#os.write(debug, "eos: start\n")
                loc.eof_stdin()
		#os.write(debug, "eos: stdin closed\n")
                loc.eof_stdout()
		#os.write(debug, "eos: stdout closed\n")
                loc.eof_stderr()
		#os.write(debug, "eos: stderr closed\n")
                rem.close()
		#os.write(debug, "eos: rem closed\n")
            elif cmd == "stdin-flow":
                loc.stdin_flow(string.atoi(data))
            else:
                raise "syntax error from server:" + str(cmd) + ":" + str(data)

	#os.write(debug, "main loop: commands performed\n")

        if loc.drained() and rem.drained():
            if exitval != None:
                if loc.epipe_seen and exitval == 0:
		    #os.write(debug, "epipe apparently seen\n")
                    sys.exit(1)
                else:
		    #os.write(debug, "exiting with %d\n" % exitval)
                    sys.exit(exitval)
            else:
                raise "don't know what to exit with"

        r0, w0, e0 = rem.select_set()
        r1, w1, e1 = loc.select_set()
	#os.write(debug, "main loop: entering select\n")
        r, w, e = select.select(r0 + r1, w0 + w1, e0 + e1, 2)
	#os.write(debug, "main loop: dealing with remote select\n")
        rem.select_action(r, w, e)
	#os.write(debug, "main loop: dealing with local select\n")
        loc.select_action(r, w, e, rem)
	#os.write(debug, "main loop: end of loop\n")
        
def main():
    ac = 1
    method = "ssh"
    if sys.argv[ac] == "-r":
	method = sys.argv[ac+1]
	ac = ac + 2
    docmd(sys.argv[ac], method, string.join(sys.argv[ac+1:]))

if __name__ == "__main__":
    #global debug
    #debug = os.dup(2)
    #debug = os.open("/dev/null", os.O_RDWR)
    #os.write(debug", "debug: writing to fd %d.\n" % debug)
    main()

==> fsh-0.2.tar.gz: fsh-0.2/fsh.py <==
#  fsh - fast remote execution
#  Copyright (C) 1999 by Per Cederqvist.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

import errno
import getopt
import getpass
import os
import select
import socket
import string
import sys

import fshlib

class remote:
    def __init__(self, method, login, server, command):
        self.s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.s.connect("/tmp/fshd-%d/%s.%s.%s" % (os.getuid(),
						  server, method, login))
        self.sock = self.s.fileno()
        self.send_queue = ""
        self.receive_queue = ""
        self.send("new %s\n" % fshlib.hollerith(command))
        
    def select_set(self):
        w = []
        if self.sock == -1:
            return [], [], []
        if self.send_queue != "":
            w.append(self.sock)
        return [self.sock], w, []

    def select_action(self, r, w, e):
        if self.sock in r:
            try:
		data = os.read(self.sock, 4096)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		else:
		    raise
	    if data == "":
                self.s.close()
                self.sock = -1
	    else:
                self.receive_queue = self.receive_queue + data

        if self.sock in w:
            try:
                sz = os.write(self.sock, self.send_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		elif eno == errno.EPIPE:
		    self.s.close()
                    self.sock = -1
		    return
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
            #os.write(debug, "sent this: %s.\n" % self.send_queue[:sz])
	    self.send_queue = self.send_queue[sz:]

    def send(self, data):
	#os.write(debug, "enqueueing this: %s.\n" % data)
        self.send_queue = self.send_queue + data
        if self.sock == -1:
            self.send_queue = ""

    def poll_response(self):
        [cmd, data], self.receive_queue = fshlib.parse_line(
            self.receive_queue, 0)

        return [cmd, data]

    def eof_seen(self):
        return self.receive_queue == "" and self.sock == -1

    def close(self):
        self.s.close()
        self.sock = -1

    def drained(self):
        return self.sock == -1

class local_side:
    def __init__(self):
        self.stdin_quota = fshlib.QUOTA
        self.stdout_quota = fshlib.QUOTA
        self.stderr_quota = fshlib.QUOTA
        self.stdin_counter = 0
        self.stdout_counter = 0
        self.stderr_counter = 0
        self.stdout_queue = ""
        self.stderr_queue = ""
        self.stdin_fd = 0
        self.stdout_fd = 1
        self.stderr_fd = 2
        self.epipe_seen = 0
	self.stdout_pending_close = 0
	self.stderr_pending_close = 0

    def select_set(self):
        r = []
        w = []
        if self.stdin_fd != -1 and self.stdin_counter < self.stdin_quota:
            r.append(self.stdin_fd)
        if self.stdout_fd != -1 and self.stdout_queue != "":
            w.append(self.stdout_fd)
        if self.stderr_fd != -1 and self.stderr_queue != "":
            w.append(self.stderr_fd)
        return r, w, []

    def select_action(self, r, w, e, srv):
        if self.stdin_fd in r:
            wanted = min(4096, self.stdin_quota - self.stdin_counter)
            assert(wanted > 0)
            try:
		data = os.read(self.stdin_fd, wanted)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		else:
		    raise
	    if data == "":
                os.close(self.stdin_fd)
                self.stdin_fd = -1
                srv.send("eof-stdin\n")
	    else:
                srv.send("stdin %s\n" % fshlib.hollerith(data))
                self.stdin_counter = self.stdin_counter + len(data)

        if self.stdout_fd in w:
	    try:
		sz = os.write(self.stdout_fd, self.stdout_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		elif eno == errno.EPIPE:
                    srv.send("eof-stdout\n")
                    os.close(self.stdout_fd)
                    self.stdout_fd = -1
		    #os.write(debug, "epipe seen I\n")
                    self.epipe_seen = 1
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
	    self.stdout_queue = self.stdout_queue[sz:]
            self.stdout_counter = self.stdout_counter + sz
	    #os.write(debug, "successful write of %d bytes.\n" % sz)
            if self.stdout_pending_close and self.stdout_queue == "":
		#os.write(debug, "closing stdout due to pend.\n")
                os.close(self.stdout_fd)
                self.stdout_fd = -1
	    #os.write(debug, "hmm.\n")
            if self.stdout_fd != -1 and \
               (self.stdout_quota - self.stdout_counter) < fshlib.QUOTA/2:

		#os.write(debug, "increasing quota.\n")
                self.stdout_quota = self.stdout_quota + fshlib.QUOTA
                srv.send("stdout-flow %s\n" % fshlib.hollerith(
                    self.stdout_quota))
		#os.write(debug, "increased quota.\n")

        if self.stderr_fd in w:
	    try:
		sz = os.write(self.stderr_fd, self.stderr_queue)
	    except os.error, (eno, emsg):
		if (eno == errno.EINTR or eno == errno.EAGAIN or eno ==
		    errno.EWOULDBLOCK):
		    return
		elif eno == errno.EPIPE:
                    srv.send("eof-stderr\n")
                    os.close(self.stderr_fd)
                    self.stderr_fd = -1
		    #os.write(debug, "epipe seen II\n")
                    self.epipe_seen = 1
		else:
		    raise
	    if sz == 0:
		raise "huh? zero-write?"
	    self.stderr_queue = self.stderr_queue[sz:]
            self.stderr_counter = self.stderr_counter + sz
            if self.stderr_pending_close and self.stderr_queue == "":
                os.close(self.stderr_fd)
                self.stderr_fd = -1
            if self.stderr_fd != -1 and \
               (self.stderr_quota - self.stderr_counter) < fshlib.QUOTA/2:

                self.stderr_quota = self.stderr_quota + fshlib.QUOTA
                srv.send("stderr-flow %s\n" % fshlib.hollerith(
                    self.stderr_quota))

    def stdout_send(self, data):
        self.stdout_queue = self.stdout_queue + data

    def stderr_send(self, data):
        self.stderr_queue = self.stderr_queue + data

    def eof_stdin(self):
        if self.stdin_fd != -1:
            os.close(self.stdin_fd)
            self.stdin_fd = -1

    def eof_stdout(self):
        self.stdout_pending_close = 1
	if self.stdout_queue == ""and self.stdout_fd != -1:
	    os.close(self.stdout_fd)
	    self.stdout_fd = -1

    def eof_stderr(self):
	#os.write(debug, "eof_stderr: start\n")
        self.stderr_pending_close = 1
	#os.write(debug, "eof_stderr: I\n")
	if self.stderr_queue == "" and self.stderr_fd != -1:
	    #os.write(debug, "eof_stderr: II (%d, %d)\n" % (debug, self.stderr_fd))
	    os.close(self.stderr_fd)
	    #os.write(debug, "eof_stderr: III\n")
	    self.stderr_fd = -1
	    #os.write(debug, "eof_stderr: IV\n")
	#os.write(debug, "eof_stderr: V\n")

    def stdin_flow(self, data):
        self.stdin_quota = data

    def drained(self):
        return self.stdout_fd == -1 and self.stderr_fd == -1

def docmd(method, login, server, command):
    rem = remote(method, login, server, command)
    loc = local_side()
    exitval = None
    while 1:
	#os.write(debug, "at top of main loop\n")
        while 1:
            cmd, data = rem.poll_response()
            if cmd == None:
                break
	    #os.write(debug, "got cmd: %s (%s)\n" % (cmd, data))
            if cmd == "stdout":
                loc.stdout_send(data)
            elif cmd == "stderr":
                loc.stderr_send(data)
            elif cmd == "eof-stdin":
                loc.eof_stdin()
            elif cmd == "eof-stdout":
                loc.eof_stdout()
            elif cmd == "eof-stderr":
                loc.eof_stderr()
            elif cmd == "exit":
                exitval = string.atoi(data)
            elif cmd == "signal-exit":
                # What do we do here?
                exitval = string.atoi(data)
            elif cmd == "eos":
		#os.write(debug, "eos: start\n")
                loc.eof_stdin()
		#os.write(debug, "eos: stdin closed\n")
                loc.eof_stdout()
		#os.write(debug, "eos: stdout closed\n")
                loc.eof_stderr()
		#os.write(debug, "eos: stderr closed\n")
                rem.close()
		#os.write(debug, "eos: rem closed\n")
            elif cmd == "stdin-flow":
                loc.stdin_flow(string.atoi(data))
            else:
                raise "syntax error from server:" + str(cmd) + ":" + str(data)

	#os.write(debug, "main loop: commands performed\n")

        if loc.drained() and rem.drained():
            if exitval != None:
                if loc.epipe_seen and exitval == 0:
		    #os.write(debug, "epipe apparently seen\n")
                    sys.exit(1)
                else:
		    #os.write(debug, "exiting with %d\n" % exitval)
                    sys.exit(exitval)
            else:
                raise "don't know what to exit with"

        r0, w0, e0 = rem.select_set()
        r1, w1, e1 = loc.select_set()
	#os.write(debug, "main loop: entering select\n")
        r, w, e = select.select(r0 + r1, w0 + w1, e0 + e1, 2)
	#os.write(debug, "main loop: dealing with remote select\n")
        rem.select_action(r, w, e)
	#os.write(debug, "main loop: dealing with local select\n")
        loc.select_action(r, w, e, rem)
	#os.write(debug, "main loop: end of loop\n")
        
def main():
    ac = 1
    method = "ssh"
    login = None
    try:
	opts, args = getopt.getopt(sys.argv[1:], "r:l:")
    except getopt.error, msg:
	sys.stderr.write(msg + "\n")
	sys.exit(1)
    if len(args) < 2:
	sys.stderr.write(
	    "fsh: usage: fsh [ -r rsh ] [ -l login ] host command [args...]\n")
	sys.exit(1)
    for opt, val in opts:
	if opt == "-r":
	    method = val
	elif opt == "-l":
	    login = val
    if login == None:
	login = getpass.getuser()
    docmd(method, login, args[0], string.join(args[1:]))

if __name__ == "__main__":
    #global debug
    #debug = os.dup(2)
    #debug = os.open("/dev/null", os.O_RDWR)
    #os.write(debug", "debug: writing to fd %d.\n" % debug)
    main()
